# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2019 DPDK Project. Developed by UNH-IOL dpdklab@iol.unh.edu.

import argparse
import atexit
import os
import shutil
import subprocess
import sys
from os import path

import paramiko

from ci_logging import logging, init_logging

ssh_cache = {}


def _close_ssh_cache():
    for client in ssh_cache:
        ssh_cache[client].close()


atexit.register(_close_ssh_cache)


class Password:
    """Mask the variable"""
    def __init__(self, var):
        self.var = var

    def get(self, mask=True):
        if mask:
            return '***'
        else:
            return self.var

    def __str__(self):
        return '***'


class CIArgParse:

    def __init__(self, description):
        self.parser = argparse.ArgumentParser(description=description)
        self.parser.add_argument(
            '--print-shell', action='store_true',
            help='Print what would be done in a shell. '
                 'This *may* not be a dry run. '
                 'Useful for sending emails.')

    def parse_args(self):
        return self.parser.parse_args()


class OS:
    CHOICES = ['ubuntu18.04', 'centos8', 'fedora31', 'opensuse-leap15',
               'archlinux']


class CI:

    def __init__(self, print_shell):
        self.print_shell = print_shell
        init_logging()

    def parse_command(self, command, mask=True):
        new_command = []
        for word in command:
            # not sure of a cleaner way
            if isinstance(word, Password):
                new_command.append(word.get(mask))
            else:
                new_command.append(word)
        return new_command

    def _print_cd(self, print_cmd, **kwargs):
        if 'cwd' in kwargs:
            logging.info('pushd ' + kwargs['cwd'])

        logging.info(print_cmd)

        if 'cwd' in kwargs:
            logging.info('popd')

    def run(self, command, exit_on_error=True, on_error=None, **kwargs):
        """Prints to output, and log to file, exit on fail."""
        print_cmd = ' '.join(self.parse_command(command))

        if self.print_shell:
            self._print_cd(print_cmd, **kwargs)
            return

        logging.info('+ ' + print_cmd)

        try:
            proc = subprocess.Popen(
                self.parse_command(command, False),
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                **kwargs)

            # live output
            for line in iter(proc.stdout.readline, b''):
                logging.info(line.decode('utf-8').strip(os.linesep))

            proc.communicate()

            if proc.returncode:
                logging.info(f'[{proc.returncode}] Error running command.')
                if on_error:
                    on_error()
                if exit_on_error:
                    sys.exit(proc.returncode)
                return proc.returncode

        except FileNotFoundError:

            logging.info(f'Command not found: {command[0]}')
            if on_error:
                on_error()
            if exit_on_error:
                sys.exit(1)

        except OSError as e:

            logging.info('Running command failed.')
            logging.info(str(e))
            if on_error:
                on_error()
            if exit_on_error:
                sys.exit(1)

    def ssh(self, host, user, password, command,
            exit_on_error=True, on_error=None):
        global ssh_cache

        if self.print_shell:
            logging.info(command)
            return

        key = host + ':' + user

        if key in ssh_cache:
            client = ssh_cache[key]
        else:
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            client.connect(host, username=user, password=password)
            ssh_cache[key] = client

        logging.info(f'Running the next command with ssh on {host}...')
        logging.info('+ ' + command)

        stdin, stdout, stderr = client.exec_command(command)

        for line in stdout:
            logging.info(line.strip(os.linesep))
        for line in stderr:
            logging.info(line.strip(os.linesep))

        return_code = stdout.channel.recv_exit_status()
        if return_code:
            logging.info(f'[{return_code}] Error running command.')
            if on_error:
                on_error()
            if exit_on_error:
                sys.exit(return_code)

    def check_call(self, command, **kwargs):
        """Prints the command to stdout and log command on print_shell"""
        print_cmd = ' '.join(command)

        if self.print_shell:
            self._print_cd(print_cmd, **kwargs)
            return

        print('+ ' + print_cmd)

        subprocess.check_call(command, **kwargs)

    def delete(self, tree):
        if path.exists(tree):
            shutil.rmtree(tree)


class ContainerMixin:

    def image_build(self, image, args=(), manager='podman'):
        script_dir = os.path.dirname(os.path.realpath(__file__))
        image = self.os + '/' + image

        # make sure the dockerfile exists
        dockerfile = script_dir + '/containers/' + image + '.Dockerfile'
        if not os.path.exists(dockerfile):
            print('Dockerfile does not exist: ' + dockerfile)
            sys.exit(1)

        # now build the image
        self.check_call((manager, 'build', '-t', image,
                         '-f', image + '.Dockerfile', '.') + args,
                        cwd=script_dir + '/containers')

    def container_run(self, image, args, on_error=None, manager='podman'):
        cmd = [manager, 'run', '--rm']

        if sys.stdout.isatty():
            cmd += ['-it']

        image = self.os + '/' + image

        self.run(cmd + args + [image], on_error=on_error)

    def kernel(self):
        if 'suse' in self.id_like:
            return 'linux-obj/x86_64/default'

        version = (subprocess.check_output(['uname', '-r'])
                   .decode('utf-8')
                   .strip('\n'))

        if 'debian' in self.id_like:
            return 'linux-headers-' + version
        elif 'arch' in self.id_like:
            return version + '/build'

        return version

    def uid(self):
        return (subprocess.check_output(['id', '-u'])
                .decode('utf-8')
                .strip('\n'))

    def gid(self):
        return (subprocess.check_output(['id', '-g'])
                .decode('utf-8')
                .strip('\n'))
