# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2019 DPDK Project. Developed by UNH-IOL dpdklab@iol.unh.edu.

import atexit
import logging.config
import os
import shutil
import sys
from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED

from utils import get_all_file_paths

OUTPUT_FOLDER = Path('out')
OUTPUT_ZIP = 'out.zip'

# This is used to initialize logging only once.
_log_init = False

LOG_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'basic': {
            'format': '%(message)s',
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'basic',
            'level': 'INFO',
        },
        'file': {
            'class': 'logging.FileHandler',
            'filename': OUTPUT_FOLDER / 'log.txt',
            'mode': 'w',
            'formatter': 'basic',
            'level': 'INFO',
        },
    },
    'root': {
        'handlers': ['console', 'file'],
        'level': 'NOTSET',
    },
}


def _save_out():
    # remove empty files and zip output folder
    file_paths = get_all_file_paths(OUTPUT_FOLDER)
    removed_files = []
    with ZipFile(OUTPUT_ZIP, 'w', ZIP_DEFLATED) as zip_file:
        for file_path in file_paths:
            if os.path.getsize(file_path) == 0:
                try:
                    os.remove(file_path)
                    removed_files.append(file_path)
                except PermissionError:
                    # windows cannot delete the log file if it gets caught
                    # in this loop when using --print-shell.
                    print('Warning: Permission error removing empty file: ' +
                          file_path)
            else:
                zip_file.write(file_path)
    if removed_files:
        print('Removed empty files: ' + ', '.join(removed_files))


def log_uncaught_exceptions(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logging.error(
        "An uncaught exception was raised. We may already know about the "
        "issue. If your patch is not run again after one business day, feel "
        "free to contact us to let us know.",
        exc_info=(exc_type, exc_value, exc_traceback))


# This function could just be called when the module is loaded,
# but I wanted to avoid creating unnecessary files when running --help.
def init_logging():
    global _log_init
    # In case multiple ci objects are created, return.
    if _log_init:
        return
    _log_init = True

    # make output folder
    if os.path.exists(OUTPUT_FOLDER):
        shutil.rmtree(OUTPUT_FOLDER)
    os.mkdir(OUTPUT_FOLDER)

    # and remove old zip
    if os.path.exists(OUTPUT_ZIP):
        os.remove(OUTPUT_ZIP)

    logging.config.dictConfig(LOG_CONFIG)
    atexit.register(_save_out)

    sys.excepthook = log_uncaught_exceptions
