#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2019 DPDK Project. Developed by UNH-IOL dpdklab@iol.unh.edu.

import multiprocessing
import os
import re
import subprocess
import time

from ci import CI, CIArgParse
from ci_logging import logging
from utils import parse_properties, get_all_file_paths


class CompileArgParse(CIArgParse):
    def __init__(self, description='DPDK compile testing'):
        super().__init__(description)
        self.parser.add_argument('tarball', help='DPDK tarball')
        # Cleaning the build is the default. --skip-clean should only be used
        # for development of the scripts.
        self.parser.add_argument(
            '--skip-clean-dpdk', action='store_true',
            help='Do not clean the dpdk build directory on startup.')
        # Building dpdk is the default. --skip-build-dpdk should only be used
        # for development of the scripts. It may be useful to use when trying
        # to just test out compiling downstream projects (ovs, spdk, etc)
        self.parser.add_argument(
            '--skip-build-dpdk', action='store_true',
            help='Skip building dpdk on startup. --skip-clean-dpdk is '
                 'assumed.')


class Compile(CI):

    def __init__(self, tarball, skip_clean_dpdk, skip_build_dpdk, **kwargs):
        super().__init__(**kwargs)

        self.tarball = tarball
        self.skip_clean_dpdk = skip_clean_dpdk
        self.skip_build_dpdk = skip_build_dpdk

        # Rather than just print() the message, log it in case these options
        # DO get used in the CI. (Someone will see the warning when they
        # download the artifacts and may contact the CI host.)
        if skip_clean_dpdk:
            logging.warning(
                'Warning: skip_clean_dpdk is set. '
                'You should not use this in the CI!')
        if skip_build_dpdk:
            logging.warning(
                'Warning: skip_build_dpdk is set. '
                'You should not use this in the CI!')

        # meant to be used in string concatenation
        self.cpu_count = str(multiprocessing.cpu_count())

        if os.name == 'nt':
            self.execenv = 'windows'
            return

        out = (subprocess.check_output(['uname', '-s'])
               .decode('utf-8')
               .strip('\n'))

        if out == 'FreeBSD':
            self.make = 'gmake'
            self.execenv = 'freebsd'
        else:
            self.make = 'make'
            self.execenv = 'linux'
            self.kernels = '/usr/src'

            with open('/etc/os-release') as f:
                properties = parse_properties(f)

            if 'ID_LIKE' in properties:
                self.id_like = properties['ID_LIKE']
            else:
                self.id_like = properties['ID']

            # use 'in' instead of equality because there can be quotes and
            # multiple values in ID_LIKE
            if 'rhel' in self.id_like or 'fedora' in self.id_like:
                self.kernels += '/kernels'
            elif 'arch' in self.id_like:
                self.kernels = '/usr/lib/modules'

    def init_dpdk(self):
        if self.skip_clean_dpdk and os.path.exists('dpdk'):
            return

        # delete previously extracted tarball
        self.delete('dpdk')

        # untar tarball (creates dpdk folder)
        try:
            self.check_call(['tar', 'xzf', self.tarball])
        except subprocess.CalledProcessError as e:
            # With bsdtar on windows, there are "can't create xyz" errors,
            # but they are not needed for the build.
            if self.execenv != 'windows':
                raise e

        if self.print_shell:
            return

        # update file times (https://github.com/ninja-build/ninja/issues/1704)
        cur_time = time.time()
        file_paths = get_all_file_paths('dpdk', return_dirs=True)
        for file_path in file_paths:
            os.utime(file_path, (cur_time, cur_time))

    def get_build_options(self):
        options = [
            'CONFIG_RTE_ENABLE_ASSERT',
            'CONFIG_RTE_EAL_ALLOW_INV_SOCKET_ID',
            'CONFIG_RTE_EAL_ALWAYS_PANIC_ON_ERROR',
            'CONFIG_RTE_ENABLE_AVX512',
            # 'CONFIG_RTE_ETHDEV_PROFILE_WITH_VTUNE', # breaks
            'CONFIG_RTE_ETHDEV_TX_PREPARE_NOOP',
        ]

        if self.execenv == 'freebsd':
            options += [
                'CONFIG_RTE_USE_LIBBSD',
            ]
        else:
            # These do not work in freebsd
            options += [
                'CONFIG_RTE_LIBRTE_MLX4_PMD',
                'CONFIG_RTE_LIBRTE_MLX5_PMD',
                'CONFIG_RTE_LIBEAL_USE_HPET',
            ]

        return options

    def apply_build_options(self):
        options = self.get_build_options()

        logging.info('Enabling compile options:\n' + '\n'.join(options))

        if self.print_shell:
            return

        with open('dpdk/config/common_base') as f:
            content = f.read()

        for option in options:
            content = re.sub(r'(' + option + r'=)n', r'\1y', content)

        with open('dpdk/config/common_base', 'w') as f:
            f.write(content)

    def build(self):
        if self.skip_build_dpdk and os.path.exists('dpdk'):
            return

        self.init_dpdk()

        self.apply_build_options()

        # configure dpdk
        self.run([self.make, 'config', f'T=x86_64-native-{self.execenv}-gcc'],
                 cwd='dpdk')

        # compile dpdk
        self.run([self.make, '-j' + self.cpu_count], cwd='dpdk')


if __name__ == '__main__':
    args = CompileArgParse().parse_args()

    compile_dpdk = Compile(**vars(args))
    compile_dpdk.build()
