#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2019 DPDK Project. Developed by UNH-IOL dpdklab@iol.unh.edu.

import os
import sys
from subprocess import check_call

from compile import Compile, CompileArgParse


class OVSCompileArgParse(CompileArgParse):
    def __init__(self):
        super().__init__('OVS Compile testing. This assumes an up-to-date ovs '
                         'git folder exists in the current directory.')
        self.parser.add_argument('--install', action='store_true',
                                 help='Whether to run "make install"')


class OVSMixin:
    def build_dpdk_and_check(self):
        if not os.path.exists('ovs'):
            print('The "ovs" folder does exist! Please make sure to clone '
                  'Open vSwitch and checkout to the "dpdk-latest" branch.')
            sys.exit(1)

        super().build()


class CompileOVS(OVSMixin, Compile):
    def __init__(self, install, **kwargs):
        super().__init__(**kwargs)

        self.install = install

    def build(self):
        self.build_dpdk_and_check()

        # create build directory
        build = os.path.join('ovs', 'build')

        if self.install:
            # running "sudo make install" changes permissions to root
            # so use root to remove
            check_call(['sudo', 'rm', '-rf', build])
        else:
            self.delete(build)

        os.mkdir(build)

        # run boot script
        self.run(['./boot.sh'], cwd='ovs')

        # run configure script
        cmd = [
            '../configure',
            '--enable-static',
            '--disable-shared',
            '--enable-ssl',
            '--with-dpdk=../../dpdk/build',
        ]

        if self.execenv != 'freebsd':
            cmd += [
                '--enable-libcapng',
            ]

        if self.install:
            cmd += [
                '--program-prefix=',
                '--disable-dependency-tracking',
                '--prefix=/usr',
                '--exec-prefix=/usr',
                '--bindir=/usr/bin',
                '--sbindir=/usr/sbin',
                '--sysconfdir=/etc',
                '--datadir=/usr/share',
                '--includedir=/usr/include',
                '--libdir=/usr/lib64',
                '--libexecdir=/usr/libexec',
                '--localstatedir=/var',
                '--sharedstatedir=/var/lib',
                '--mandir=/usr/share/man',
                '--infodir=/usr/share/info'
            ]

        self.run(cmd, cwd=build)

        # compile ovs
        self.run([self.make, '-j' + self.cpu_count], cwd=build)

        if self.install:
            self.run(['sudo', self.make, 'install'], cwd=build)


if __name__ == '__main__':
    args = OVSCompileArgParse().parse_args()

    compile_ovs = CompileOVS(**vars(args))
    compile_ovs.build()
