#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2019 DPDK Project. Developed by UNH-IOL dpdklab@iol.unh.edu.

import os
import subprocess
import sys
import time

from ci import CI, Password, CIArgParse
from ci_logging import logging, OUTPUT_FOLDER


class OVSArgParse(CIArgParse):
    def __init__(self):
        super().__init__(
            'OVS Performance testing. This assumes compile_ovs was already '
            'ran. This is expected to run on the tester.')
        self.parser.add_argument(
            '--nic', required=True, help='The nic to test')
        self.parser.add_argument(
            '--nic-driver', help='The original nic driver (optional)')
        self.parser.add_argument(
            '--virt-nic-driver',
            help='The virtual nic driver to switch (optional)')
        self.parser.add_argument(
            '--trex', required=True,
            help='The root trex directory')
        self.parser.add_argument(
            '--cfg', required=True,
            help='The trex config file to use')
        self.parser.add_argument(
            '--ovs-address', required=True,
            help='The ovs host address')
        self.parser.add_argument(
            '--ovs-user', required=True,
            help='The ovs dut user')
        self.parser.add_argument(
            '--ovs-password', required=True,
            help='The ovs dut password')
        self.parser.add_argument(
            '--basic', action='store_true',
            help='Don\'t run the full test suite')


class RunOVS(CI):
    def __init__(self, nic, nic_driver, virt_nic_driver, trex, cfg,
                 ovs_address, ovs_user, ovs_password, basic, **kwargs):
        super().__init__(**kwargs)

        self.nic = nic
        self.nic_driver = nic_driver
        self.virt_nic_driver = virt_nic_driver
        self.trex = trex
        self.cfg = cfg
        self.ovs_address = ovs_address
        self.ovs_user = ovs_user
        self.ovs_password = ovs_password
        self.basic = basic

    def ssh(self, command, **kwargs):
        super().ssh(self.ovs_address, self.ovs_user, self.ovs_password,
                    command, **kwargs)

    def revert(self):
        self.ssh('sudo journalctl -n 20')
        if self.nic_driver:
            self.ssh(f'sudo dpdk-devbind -b {self.nic_driver} {self.nic}')

    def destroy(self):
        self.ssh('sudo virsh destroy rhel_loopback')

    def clean_up(self):
        # we've had problems in the past where the vm would not shutdown
        self.ssh('sudo virsh shutdown rhel_loopback', on_error=self.destroy,
                 exit_on_error=False)

        self.ssh('sudo systemctl stop openvswitch-switch')

        # put back the driver to normal
        if self.nic_driver:
            self.ssh(f'sudo dpdk-devbind -b {self.nic_driver} {self.nic}')

    def run_ovs(self, flow_type, output, exit_on_error=True):
        os.mkdir(output)

        cmd = ['/opt/ovs_perf/ovs_performance.py',
               '-d',  # The debug flag is required - the test breaks otherwise
               '-l', 'testrun_log.txt',
               '--tester-type', 'trex',
               '--tester-address', 'localhost',
               '--tester-interface', '0',
               '--ovs-address', self.ovs_address,
               '--ovs-user', self.ovs_user,
               '--ovs-password', Password(self.ovs_password),
               '--dut-vm-address', '192.168.122.62',
               '--dut-vm-user', 'root',
               '--dut-vm-password', Password('root'),
               '--dut-vm-nic-queues', '2',
               '--physical-interface', 'dpdk0',
               '--physical-speed', '10',
               '--virtual-interface', 'vhost0',
               '--dut-vm-nic-pci', '0000:00:02.0',
               '--no-bridge-config',
               '--skip-pv-test']

        if self.basic:
            cmd += ['--packet-list', '64', '--stream-list', '1000']
        else:
            cmd += [
               '--flow-type', flow_type,
               '--run-time', '100',
               '--stream-list', '1000,100000',
               '--packet-list', '64,512,1512',
            ]

        return self.run(cmd, exit_on_error, cwd=output)

    def run_perf(self):
        # just make sure the vm is not running
        self.ssh('sudo virsh shutdown rhel_loopback', exit_on_error=False)

        # we've had permission errors from libvirt, which were fixed by
        # restarting libvirtd
        self.ssh('sudo systemctl restart libvirtd')

        if self.virt_nic_driver:
            self.ssh(f'sudo dpdk-devbind -b {self.virt_nic_driver} {self.nic}')

        # we had errors in the past with this starting up, so also log it
        self.ssh('sudo systemctl start openvswitch-switch',
                 on_error=self.revert)

        # set up the ports on the nic.
        # for a system with one nic, this was not necessary, but since we have
        # systems with multiple nics, and the test suite requires the nic
        # be on dpdk0, we set up the nics anyways.
        self.ssh('sudo ovs-vsctl del-port ovs_pvp_br0 dpdk0',
                 exit_on_error=False)
        self.ssh('sudo ovs-vsctl add-port ovs_pvp_br0 dpdk0 -- '
                 'set Interface dpdk0 type=dpdk -- '
                 'set Interface dpdk0 options:dpdk-devargs=' + self.nic +
                 ' -- '
                 'set interface dpdk0 options:n_rxq=2 '
                 'other_config:pmd-rxq-affinity="0:1,1:15" -- '
                 'set Interface dpdk0 ofport_request=1')

        self.ssh('sudo virsh start rhel_loopback')

        trex_cmd = ['sudo', './t-rex-64', '-c', '4', '-i',
                    '--no-scapy-server', '--cfg', self.cfg]
        logging.info('+ ' + ' '.join(trex_cmd))
        trex = subprocess.Popen(trex_cmd, cwd=self.trex,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        # make sure trex is fully running (otherwise ovs perf will fail)
        time.sleep(30)

        logging.info('Checking if t-rex is running...')

        try:
            stdout, stderr = trex.communicate(timeout=10)
            if trex.returncode or 'ERROR' in stdout:
                logging.info('t-rex failed to start!')
                logging.info(stdout.decode('utf-8'))
                self.clean_up()
                sys.exit(trex.returncode or 1)
        except subprocess.TimeoutExpired:
            # It means that t-rex is running
            pass

        logging.info('t-rex started!')

        self.ssh('ovs-vswitchd --version')

        # run testing
        # If an error happens, run the rest of the script to clean up
        # everything (on_error could have been used, but I arbitrarily
        # chose not to use it)
        code = self.run_ovs('L2', OUTPUT_FOLDER / 'resultsL2', False)
        # run l3 if l2 is successful
        if not code:
            code = self.run_ovs('L3', OUTPUT_FOLDER / 'resultsL3', False)

        # can't use terminate() because sudo was used
        subprocess.check_call(['sudo', 'kill', '-SIGTERM', str(trex.pid)])

        self.clean_up()

        return code


if __name__ == '__main__':
    args = OVSArgParse().parse_args()

    run_ovs = RunOVS(**vars(args))
    ret = run_ovs.run_perf()
    if ret:
        sys.exit(ret)
